from pathlib import Path

import pytest
import yaml
from p4p.client.thread import Context

from p4pillon.thread.config_reader import parse_config
from p4pillon.thread.server import SimpleServer

root_dir = Path(__file__).parents[2]


@pytest.fixture
def ntscalar_config():
    with open(f"{root_dir}/integration/ntscalar_config.yml") as f:
        ntscalar_dict = yaml.load(f, Loader=yaml.SafeLoader)
        f.close()
    return ntscalar_dict


@pytest.fixture
def ntenum_config():
    with open(f"{root_dir}/integration/ntenum_config.yml") as f:
        ntenum_dict = yaml.load(f, Loader=yaml.SafeLoader)
        f.close()
    return ntenum_dict


@pytest.fixture()
def enum_yaml_server(ntenum_config) -> SimpleServer:
    """fixture that handles creation and desctruction of server for use during tests"""
    # NOTE this could be tidied up by moving the tests into classes and using
    # setupclass and teardownclass methods as suggested here:
    # https://docs.pytest.org/en/latest/how-to/xunit_setup.html
    # NOTE also that this fixture has to be used as a parameter in every test
    # for the server to actually run
    server = start_server(ntenum_config)
    yield server
    server.stop()


@pytest.fixture()
def yaml_server(ntscalar_config) -> SimpleServer:
    """fixture that handles creation and desctruction of server for use during tests"""
    # NOTE this could be tidied up by moving the tests into classes and using
    # setupclass and teardownclass methods as suggested here:
    # https://docs.pytest.org/en/latest/how-to/xunit_setup.html
    # NOTE also that this fixture has to be used as a parameter in every test
    # for the server to actually run
    server = start_server(ntscalar_config)
    yield server
    server.stop()


@pytest.fixture()
def basic_server() -> SimpleServer:
    server = SimpleServer(
        prefix="TEST:",
    )
    yield server
    server.stop()


@pytest.fixture()
def ctx() -> Context:
    client_context = Context("pva")
    yield client_context
    client_context.close()


def start_server(config: dict):
    # NOTE this will be replaced by a more universal `parse_yaml` function or equivalent
    server = SimpleServer(
        prefix="TEST:",
    )
    parse_config(config, server)

    server.start()
    return server



"""
Patch in p4p PR #172 - support for open(), post(), and close() handler functions
"""

import logging

from p4p.server.raw import SharedPV as _SharedPV

_log = logging.getLogger(__name__)


class Handler:
    """Skeleton of SharedPV Handler

    Use of this as a base class is optional.

    This is an alternative handler with added open(), post(), and close() to the set of functions implemented in p4p.

    """

    def open(self, value):
        """
        Called each time an Open operation is performed on this Channel

        :param value:  A Value, or appropriate object (see nt= and wrap= of the constructor).
        """
        pass

    def put(self, pv, op):
        """
        Called each time a client issues a Put
        operation on this Channel.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        :param ServerOperation op: The operation being initiated.
        """
        op.done(error="Not supported")

    def post(self, pv, value):
        """
        Called each time a client issues a post
        operation on this Channel.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        :param value:  A Value, or appropriate object (see nt= and wrap= of the constructor).
        :param dict options: A dictionary of configuration options.
        """
        pass

    def rpc(self, pv, op):
        """
        Called each time a client issues a Remote Procedure Call
        operation on this Channel.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        :param ServerOperation op: The operation being initiated.
        """
        op.done(error="Not supported")

    def onFirstConnect(self, pv):
        """
        Called when the first Client channel is created.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        """
        pass

    def onLastDisconnect(self, pv):
        """
        Called when the last Client channel is closed.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        """
        pass

    def close(self, pv):
        """
        Called when the Channel is closed.

        :param SharedPV pv: The :py:class:`SharedPV` which this Handler is associated with.
        """
        pass


class SharedPV(_SharedPV):
    """Shared state Process Variable.  Callback based implementation.

    .. note:: if initial=None, the PV is initially **closed** and
              must be :py:meth:`open()`'d before any access is possible.

    :param handler: A object which will receive callbacks when eg. a Put operation is requested.
                    May be omitted if the decorator syntax is used.
    :param Value initial: An initial Value for this PV.  If omitted, :py:meth:`open()`s must be called before client access is possible.
    :param nt: An object with methods wrap() and unwrap().  eg :py:class:`p4p.nt.NTScalar`.
    :param callable wrap: As an alternative to providing 'nt=', A callable to transform Values passed to open() and post().
    :param callable unwrap: As an alternative to providing 'nt=', A callable to transform Values returned Operations in Put/RPC handlers.
    :param dict options: A dictionary of configuration options.

    Creating a PV in the open state, with no handler for Put or RPC (attempts will error). ::

        from p4p.nt import NTScalar
        pv = SharedPV(nt=NTScalar('d'), value=0.0)
        # ... later
        pv.post(1.0)

    The full form of a handler object is: ::

        class MyHandler:
            def put(self, pv, op):
                pass
            def rpc(self, pv, op):
                pass
            def onFirstConnect(self): # may be omitted
                pass
            def onLastDisconnect(self): # may be omitted
                pass
    pv = SharedPV(MyHandler())

    Alternatively, decorators may be used. ::

        pv = SharedPV()
        @pv.put
        def onPut(pv, op):
            pass

    The nt= or wrap= and unwrap= arguments can be used as a convience to allow
    the open(), post(), and associated Operation.value() to be automatically
    transform to/from :py:class:`Value` and more convienent Python types.
    See :ref:`unwrap`
    """

    def open(self, value, nt=None, wrap=None, unwrap=None, **kws):
        """Mark the PV as opened an provide its initial value.
        This initial value is later updated with post().

        :param value:  A Value, or appropriate object (see nt= and wrap= of the constructor).

        Any clients which have begun connecting which began connecting while
        this PV was in the close'd state will complete connecting.

        Only those fields of the value which are marked as changed will be stored.
        """

        self._wrap = wrap or (nt and nt.wrap) or self._wrap
        self._unwrap = unwrap or (nt and nt.unwrap) or self._unwrap

        try:
            V = self._wrap(value, **kws)
        except Exception as exc:  # py3 will chain automatically, py2 won't
            raise ValueError(f"Unable to wrap {value} with {self._wrap} and {kws}") from exc

        # Guard goes here because we can have handlers that don't inherit from
        # the Handler base class
        try:
            open_fn = self._handler.open
        except AttributeError:
            pass
        else:
            open_fn(V)

        _SharedPV.open(self, V)

    def post(self, value, **kws):
        """Provide an update to the Value of this PV.

        :param value:  A Value, or appropriate object (see nt= and wrap= of the constructor).

        Only those fields of the value which are marked as changed will be stored.

        Any keyword arguments are forwarded to the NT wrap() method (if applicable).
        Common arguments include: timestamp= , severity= , and message= .
        """
        try:
            V = self._wrap(value, **kws)
        except Exception as exc:  # py3 will chain automatically, py2 won't
            raise ValueError(f"Unable to wrap {value} with {self._wrap} and {kws}") from exc

        # Guard goes here because we can have handlers that don't inherit from
        # the Handler base class
        try:
            post_fn = self._handler.post
        except AttributeError:
            pass
        else:
            post_fn(self, V)

        _SharedPV.post(self, V)

    def close(self, destroy=False):
        """Close PV, disconnecting any clients.
        :param bool destroy: Indicate "permanent" closure.  Current clients will not see subsequent open().
        close() with destory=True or sync=True will not prevent clients from re-connecting.
        New clients may prevent sync=True from succeeding.
        Prevent reconnection by __first__ stopping the Server, removing with :py:meth:`StaticProvider.remove()`,
        or preventing a :py:class:`DynamicProvider` from making new channels to this SharedPV.
        """
        try:
            close_fn = self._handler.close
        except AttributeError:
            pass
        else:
            close_fn(self)

        _SharedPV.close(self)

    class _WrapHandler(_SharedPV._WrapHandler):  # pylint: disable=W0212
        "Wrapper around user Handler which logs exceptions"

        def open(self, value):
            _log.debug("OPEN %s %s", self._pv, value)
            try:
                self._pv._exec(None, self._real.open, value)
            except AttributeError:
                pass

        def post(self, value):
            _log.debug("POST %s %s", self._pv, value)
            try:
                self._pv._exec(None, self._real.post, self._pv, value)
            except AttributeError:
                pass

        def close(self):
            _log.debug("CLOSE %s", self._pv)
            try:
                self._pv._exec(None, self._real.close, self._pv)
            except AttributeError:
                pass



import numpy as np
from p4p.client.thread import Context

from p4pillon.definitions import MAX_FLOAT, MAX_INT32, MIN_FLOAT, MIN_INT32, AlarmSeverity, Format


def assert_value_changed(pvname: str, put_value, put_timestamp: float, ctx: Context):
    pv_state = ctx.get(pvname)
    current_value = pv_state.raw.todict()["value"]
    assert np.array_equal(np.array(current_value), np.array(put_value))
    # assert pv_state.timestamp >= put_timestamp  # TODO: Check why this timestamp is broken?


def assert_value_not_changed(pvname: str, put_value, ctx: Context):
    pv_state = ctx.get(pvname)
    current_value = pv_state.raw.todict()["value"]
    assert not np.array_equal(np.array(current_value), np.array(put_value))


def assert_alarm_present(ctx: Context, pvname: str):
    pv_state = ctx.get(pvname)

    for key in ["severity", "status", "message"]:
        assert pv_state.raw.todict().get("alarm").get(key) is not None


def assert_correct_display_config(pv_state: dict, pv_config: dict):
    display_state = pv_state.get("display")
    display_config = pv_config.get("display")
    if display_config is None:
        # this occurs when the entry in the yaml file is present but nothing
        # listed within it
        display_config = {}

    assert display_state.get("description") == pv_config.get("description")
    assert display_state.get("units") == display_config.get("units", "")
    assert display_state.get("form") == {
        "index": Format[display_config.get("format", "DEFAULT")].value[0],
        "choices": [form.value[1] for form in Format],
    }

    assert display_state.get("precision") == display_config.get("precision", 2)
    assert display_state.get("limitHigh") == display_config.get("high", MAX_FLOAT)
    assert display_state.get("limitLow") == display_config.get("low", MIN_FLOAT)


def assert_correct_control_config(pv_state: dict, pv_config: dict):
    control_state = pv_state.get("control")
    control_config = pv_config.get("control")
    if control_config is None:
        # this occurs when the entry in the yaml file is present but nothing
        # listed within it
        control_config = {}

    if pv_config["type"] == "DOUBLE":
        default_max, default_min = MAX_FLOAT, MIN_FLOAT
    else:
        default_max, default_min = MAX_INT32, MIN_INT32

    assert control_state.get("limitLow") == control_config.get("low", default_min)
    assert control_state.get("limitHigh") == control_config.get("high", default_max)
    assert control_state.get("minStep") == control_config.get("min_step", 0)


def assert_correct_alarm_config(pv_state: dict, pv_config: dict):
    valueAlarm_state = pv_state.get("valueAlarm")

    valueAlarm_config = pv_config.get("valueAlarm")
    if valueAlarm_config is None:
        # this occurs when the entry in the yaml file is present but nothing
        # listed within it
        valueAlarm_config = {}

    if pv_config["type"] == "DOUBLE":
        default_max, default_min = MAX_FLOAT, MIN_FLOAT
    else:
        default_max, default_min = MAX_INT32, MIN_INT32

    assert valueAlarm_state.get("lowAlarmLimit") == valueAlarm_config.get("low_alarm", default_min)
    assert valueAlarm_state.get("lowWarningLimit") == valueAlarm_config.get("low_warning", default_min)
    assert valueAlarm_state.get("highAlarmLimit") == valueAlarm_config.get("high_alarm", default_max)
    assert valueAlarm_state.get("highWarningLimit") == valueAlarm_config.get("high_warning", default_max)
    assert valueAlarm_state.get("lowAlarmSeverity") == AlarmSeverity.MAJOR_ALARM.value
    assert valueAlarm_state.get("lowWarningSeverity") == AlarmSeverity.MINOR_ALARM.value
    assert valueAlarm_state.get("highAlarmSeverity") == AlarmSeverity.MAJOR_ALARM.value
    assert valueAlarm_state.get("highWarningSeverity") == AlarmSeverity.MINOR_ALARM.value
    assert valueAlarm_state.get("hysteresis") == 0


def assert_pv_in_major_alarm_state(pvname: str, ctx: Context):
    val = ctx.get(pvname)
    assert val.severity == AlarmSeverity.MAJOR_ALARM.value


def assert_pv_in_minor_alarm_state(pvname: str, ctx: Context):
    val = ctx.get(pvname)
    assert val.severity == AlarmSeverity.MINOR_ALARM.value


def assert_pv_in_invalid_alarm_state(pvname: str, ctx: Context):
    val = ctx.get(pvname)
    assert val.severity == AlarmSeverity.INVALID_ALARM.value


def assert_pv_not_in_alarm_state(pvname: str, ctx: Context):
    val = ctx.get(pvname)
    assert val.severity == AlarmSeverity.NO_ALARM.value


def assert_enum_value_changed(pvname: str, put_value: dict, put_timestamp: float, ctx: Context):
    pv_state = ctx.get(pvname)
    current_value = pv_state.raw.todict()["value"]
    assert put_value == current_value
    print(
        f"time from PV = {pv_state.timestamp}, time at put = {put_timestamp}, diff = {pv_state.timestamp - put_timestamp}"
    )
    assert pv_state.timestamp >= put_timestamp


def assert_enum_value_not_changed(pvname: str, put_value, ctx: Context):
    pv_state = ctx.get(pvname)
    current_value = pv_state.raw.todict()["value"]
    assert put_value != current_value



from abc import ABC, abstractmethod

from p4pillon.pvrecipe import BasePVRecipe
from p4pillon.server.raw import SharedPV


class BaseSimpleServer(ABC):
    """
    Creates PVs and manages their lifetimes
    """

    @abstractmethod
    def __init__(self, prefix: str = "") -> None:
        """
        Initialize the SimpleServer instance.

        :param prefix: The prefix to be added to the PVs (Process Variables) of the server e.g. DEV: Defaults to "".
        """
        # the prefix determines the prefix of the PVs to be added to the server e.g. DEV:
        self.prefix: str = prefix  #: The prefix to be prepended to the PVs generated by this server.
        self._pvs: dict[str, SharedPV] = {}

    @abstractmethod
    def add_pv(self, pv_name: str, pv_recipe: BasePVRecipe) -> SharedPV:
        """
        Add a PV to the server

        :param pv_name: The name of the PV to be added.
        :param pv_recipe: The recipe with instructions for creating the PV.
        :return: The created PV.
        """

    @abstractmethod
    def remove_pv(self, pv_name: str) -> None:
        """
        Remove a PV from the server

        :param pv_name: The name of the PV to be removed.
        :raises KeyError: If the PV is not found in the list managed by the server.
        """

    @property
    def pvlist(self) -> list[str]:
        """The PVs managed by the server"""
        return list(self._pvs.keys())

    def __getitem__(self, pv_name: str) -> SharedPV | None:
        """Return one of the PVs managed by the server given its name"""
        if not pv_name.startswith(self.prefix):
            pv_name = self.prefix + pv_name
        return self._pvs.get(pv_name)
